

/************************** Interrupt handlers ***************************/

/**********************interrupt for buttons 1 and 3**********************/

void button_press_13 (void)
{
    // allow interrupts while we're doing this
    //sei();
    if (! (PIND & _BV(BUTTON1))) 
        {
        // button1 is pressed
        if (! (last_buttonstate & 0x1)) 
            { // was not pressed before
            PCMSK2 &= ~_BV(PCINT21);        //its better to disable the interrupt here to avoid resets
            delayms(50);                    // debounce
            PCMSK2 |= _BV(PCINT21);         //enable interrupt again
            if (PIND & _BV(BUTTON1))        // filter out bounces
                return;
            // reset nightmode timer
            nightmodetimer = NIGHTMODEONTIME;
            // check if we will snag this button press for alarm off
            if (ctrl_struct.alarming) 
                {
                // turn off alarm
                alarm_off();
                setsnooze();
                ctrl_struct.alarming = FALSE;
                return;
                }
            else if (ctrl_struct.countdown_alarming) 
                {
                alarm_off();
                setsnooze();
                return;
                }
            else if (ctrl_struct.nightmode == TRUE)
                {
                releasenightmode();
                return;
                }
            last_buttonstate |= 0x1;
            just_pressed |= 0x1;
            }
        } 
    else 
        last_buttonstate &= ~0x1;

    if (! (PIND & _BV(BUTTON3))) 
        {
        // button3 is pressed
        if (! (last_buttonstate & 0x4)) 
            { // was not pressed before
            PCMSK2 &= ~_BV(PCINT20);        //its better to disable the interrupt here to avoid resets
            delayms(50);                    // debounce
            PCMSK2 |= _BV(PCINT20);         //enable interrupt again
            if (PIND & _BV(BUTTON3))        // filter out bounces
                return;
            buttonholdcounter = 2;          // see if we're press-and-holding
            while (buttonholdcounter) 
                {
                if (PIND & _BV(BUTTON3)) 
                    {        // released
                    last_buttonstate &= ~0x4;
                    // reset nightmode timer
                    nightmodetimer = NIGHTMODEONTIME;
                    // check if we will snag this button press for alarm off
                    if (ctrl_struct.alarming) 
                        {
                        // turn off alarm
			alarm_off();
                        setsnooze();
	   		return;
                        }
                    else if (ctrl_struct.countdown_alarming) 
                        {
			ctrl_struct.countdown_alarming = FALSE;
			alarm_off();
			return;
                        } 
                    else if (ctrl_struct.nightmode == TRUE)
                        {
                        releasenightmode();
                        return;
                        }
                    just_pressed |= 0x4;
                    return;
                    } 
                }
            last_buttonstate |= 0x4;
            pressed |= 0x4;                 // held down
            }
        } 
    else 
        {
        pressed = 0;                      // button released
        last_buttonstate &= ~0x4;
        }
}

/*************************Interrupt for button #2****************************/

void button_press_2(void)
{
    if (! (PINB & _BV(BUTTON2))) 
        {
        // button2 is pressed
    
        if (! (last_buttonstate & 0x2)) 
            { // was not pressed before
            PCMSK0 &= ~_BV(PCINT0);        //its better to disable the interrupt here to avoid resets
            delayms(50);                    // debounce
            PCMSK0 |= _BV(PCINT0);         //enable interrupt again
            if (PINB & _BV(BUTTON2))        // filter out bounces
                return;
            // reset nightmode timer
            nightmodetimer = NIGHTMODEONTIME;
            // check if we will snag this button press for alarm off
            if (ctrl_struct.alarming) 
                {
                ctrl_struct.alarming = FALSE;
		alarm_off();
		return;
                }
            else if (ctrl_struct.countdown_alarming) 
                {
		ctrl_struct.countdown_alarming = FALSE;
		alarm_off();
		return;
                }
            else if (ctrl_struct.nightmode == TRUE)
                {
                releasenightmode();
                return;
                }
            else if (displaymode == SHOW_TIME) 
                {
                if (ctrl_struct.alarm_on)
                    {
                    setalarmstate(FALSE);
                    }
                else
                    {
                    setalarmstate(TRUE);
                    }
                }
            last_buttonstate |= 0x2;
            just_pressed |= 0x2;
            }
        } 
    else 
    last_buttonstate &= ~0x2;
}

/***********************Interrupt for DCF**********************/

void dcf_analysis(void)
{
    if (ctrl_struct.dcf_sync)
        {
        display3[0] |= 0x2;
        return;
        }
    sei();
    if (!(DCF_PIN & _BV(DCF)))
	{
        if(dcf_struct.counter_high>=1600 && dcf_struct.counter_high<=2000)    //minute sync
            {
                if(dcf_struct.bit_number==59 && dcf_struct.minute_parity%2==0 && dcf_struct.hour_parity%2==0 && dcf_struct.date_parity%2==0)
                    {
                    dcf_struct.bit_number=0;
                    dcf_struct.counter_high=0;
                    // we will check day,month, year, minute and hour against values from last minute before updating the clock to avoid failures in dcf sync
                    if (dcf_struct.day == dcf_struct.day_check && dcf_struct.month == dcf_struct.month_check && dcf_struct.year == dcf_struct.year_check && dcf_struct.minute == dcf_struct.minute_check + 1 && dcf_struct.hour == dcf_struct.hour_check)
                        {
                        dcf_match_date();
                        dcf_match_time();
                        if (effect_ctrl.light == FALSE)
                            recreate_pwm_sequence();
                        }
                    else
                        {
                        dcf_struct.bit_number=0;
                        dcf_struct.counter_high=0;
                        dcf_store_check();
                        }
                    }
                else
                    {
                        dcf_struct.bit_number=0;
                        dcf_struct.counter_high=0;
                        dcf_date_reset();
                        dcf_time_reset();
                    }
            }
        dcf_struct.counter_low=0;
	}
    else
	{
        if(dcf_struct.counter_low>=60 && dcf_struct.counter_low<=120)	//log 0
            {
                dcf_0();
                tick();
                dcf_struct.bit_number++;
                dcf_struct.counter_low=0;
                dcf_struct.counter_high=0;
            }
        else if(dcf_struct.counter_low>=160 && dcf_struct.counter_low<=220)	//log 1
            {	
                dcf_1();
                tick();
                dcf_struct.bit_number++;
                dcf_struct.counter_low=0;
                dcf_struct.counter_high=0;
            }
        }
}

/***************Comparator interrupt for detect low-power-mode**********/

void set_power_mode(void)
{
    //power down -> main power lost
    if (ctrl_struct.sleepmode == FALSE) 
        {
        //save the time, always good in power cycling and before doing reset with app_start.
        save_time();
        //prepare standby
        gotosleep();
        //restart
        app_start();
        }
    //main power restored
    else 
        {
        //save the time, always good in power cycling and before doing reset with app_start.
        save_time();
        app_start();
        }
} 

/**************Calculate date, time and other stuff once a second*************/

void time_calc(void)
{

    static uint8_t time_divider;
    uint8_t nightmode_time;
   
    //call ambilight_pwm 4 times a second
    ambilight_pwm_calc();

    time_divider++;
    if (time_divider < 4)
	{
	return;
	}

    //from here with 1 Hz
    time_divider = 0;
    
    CLKPR = _BV(CLKPCE);  //MEME
    CLKPR = 0;
    time.s++;             // one second has gone by
    overblend = 200;	// new overblending every second;

    // read temperature every 10 seconds if we are not in sleep mode
    if (ctrl_struct.sleepmode == FALSE && !(time.s % 10))
            read_temperature();
    
    // a minute!
    if (time.s >= 60) 
        {
        time.s = 0;
        time.m++;
    
        // Reset all floating effects so that they will come once a minute
        effect_ctrl.date = TRUE;
        effect_ctrl.message_timer = TRUE;
        effect_ctrl.time = TRUE;
        effect_ctrl.temp = TRUE;

        // decrease countdown timer
        if (countdown && ctrl_struct.countdown_is_set)
            countdown--;
        
        // decrease alarm timer
        if (alarmtimer && ctrl_struct.alarming)
            alarmtimer--;        
        }

    // an hour...
    if (time.m >= 60) 
        {
        time.m = 0;
        time.h++; 
        if (ctrl_struct.dst == TRUE)
            calc_timezone();

        //we should store values in eeprom now (its no problem - its possible to write to eeprom 100.000 times)
        save_time();
        }

    // a day....
    if (time.h >= 24) 
        {
        time.h = 0;
        time.d++;

        //sync pwm to prevent mismatch between color and clock
        if (effect_ctrl.light == FALSE)
            recreate_pwm_sequence();
    
        //redo dcf sync, setting the check variables will make sync 1 min faster
        dcf_struct.day_check=time.d;
        dcf_struct.month_check=time.mo;
        dcf_struct.year_check=time.y;
        dcf_struct.minute_check=time.m;
        dcf_struct.hour_check=time.h;
        ctrl_struct.dcf_sync = FALSE;
    
        //if time.z is true, its switched to wintertime. we have to set it to false again next day. for details see util.c
        if (time.z == TRUE && ctrl_struct.dst == TRUE)
            time.z = FALSE;
    }

    // a full month!
    // we check the leapyear and date to verify when its time to roll over months

    if ((time.d > 31) ||
      ((time.d == 31) && ((time.mo == 4)||(time.mo == 6)||(time.mo == 9)||(time.mo == 11))) ||
      ((time.d == 30) && (time.mo == 2)) ||
      ((time.d == 29) && (time.mo == 2) && !leapyear(2000+time.y))) 
        {
        time.d = 1;
        time.mo++;
        }
  
    // HAPPY NEW YEAR!
    if (time.mo >= 13) 
        {
        time.y++;
        time.mo = 1;
        }
  
    // If we're in low power mode we should get out now since the display is off
    if (ctrl_struct.sleepmode)
        return;
   
    // update clock display if we are not in nightmode
    if (displaymode == SHOW_TIME && ctrl_struct.nightmode == FALSE) 
	control_display();

    if ((ctrl_struct.alarm_on && (alarm_h == time.h) && (alarm_m == time.m) && (time.s == 0)))  
        {
    	ctrl_struct.alarming = TRUE;
        snoozetimer = 0;
        alarmtimer = MAXALARMTIMER;
	}
    else if (ctrl_struct.alarming && alarmtimer == 0)
        {
        ctrl_struct.alarming = FALSE;
        alarm_off();
        }
    else if (ctrl_struct.countdown_is_set && countdown == 0)  
        {
	ctrl_struct.countdown_is_set = FALSE;
	ctrl_struct.countdown_alarming = TRUE;
  	}

    // decrease nightmodetimer and start nightmode if nightmode is enabled, its the right time and we are not alarming,in menu or with wrong time. Otherwise release from night mode.
    // if (nightmode_on_hour == 0 && nightmode_off_hour == 0) is true, nightmode is disabled.
    if (!(nightmode_on_hour == 0 && nightmode_off_hour == 0))
        {
        //detect if we are within nightmode time, covers both cases:starthour greater than endhour (ex 05 - 17) or starthour smaller/equal than endhour (example 23 - 05)
        if ((nightmode_on_hour < nightmode_off_hour && time.h >= nightmode_on_hour && time.h < nightmode_off_hour) || (nightmode_on_hour >= nightmode_off_hour && (time.h >= nightmode_on_hour || time.h < nightmode_off_hour)))
            nightmode_time = TRUE;
        else
            nightmode_time = FALSE;
        if (ctrl_struct.nightmode == FALSE && nightmode_time == TRUE && nightmodetimer > 0 && displaymode == SHOW_TIME && ctrl_struct.alarming == FALSE && ctrl_struct.countdown_alarming == FALSE && ctrl_struct.timeunknown == FALSE)
            {
            nightmodetimer--;
            if (nightmodetimer == 0)
                gotonightmode();
            }
        else if ((ctrl_struct.nightmode == TRUE && nightmode_time == FALSE) || (ctrl_struct.alarming == TRUE || ctrl_struct.countdown_alarming == TRUE || ctrl_struct.timeunknown == TRUE))
            releasenightmode();
        }
    if (timeoutcounter)
        timeoutcounter--;
    if (buttonholdcounter)
        buttonholdcounter--;
    if (snoozetimer) 
        {
        snoozetimer--;
        if (snoozetimer % 2) 
            display[0] |= 0x1;
        else
            display[0] &= ~0x1;
        }
}

/*******Update Display and Alarming * called @ ~30khz (31.25 khz)*********/

void update_display(void)
{
    
    // allow other interrupts to go off while we're doing display updates
    sei();

    static uint8_t pwm_div;

    //use ambilight pwm if light is switched to "puls" in menu
    //its called via devider,otherwise its much for this little mcu ;(
    
    if (ctrl_struct.nightmode == FALSE)
        {
        pwm_div++;
        if (pwm_div == 3)
            {
            pwm_div = 0;
            ambilight_pwm();
            }
        }
        
    // kick the dog
    kickthedog();

    // divide down to 100Hz * digits
    muxdiv++;
    if (muxdiv < MUX_DIVIDER)
        return;
    muxdiv = 0;
    // now at 100Hz * digits
    if (dcf_struct.counter_low < 65534)
        dcf_struct.counter_low++;	//Counter for DCF low
    if (dcf_struct.counter_high < 65534)
        dcf_struct.counter_high++;	//Counter for DCF high
    if (milliseconds < 65534)
        milliseconds++;

    //update display if we are not in night mode
    if (ctrl_struct.nightmode == FALSE)
        {
        // Cycle through each digit in the display
        if (currdigit >= DISPLAYSIZE)
            currdigit = 0;

        // This creates a shadow of the last display for overblending in time-mode 
        if (overblend && effect_ctrl.create_shadow_en == TRUE && effect_ctrl.effect == TRUE)
            {
            overblend--;
            setdisplay(currdigit, display2[currdigit]);
            }

        // Set the current display's segments
        setdisplay(currdigit, display[currdigit]);
        // and go to the next
        currdigit++;
        }

    // check if we should have the alarm on
    if ((ctrl_struct.alarming && !snoozetimer) || ctrl_struct.countdown_alarming) 
        {
	alarmdiv++;
    	if (alarmdiv > ALARM_DIVIDER) 
            alarmdiv = 0;
    	else
            return;
        // This part only gets reached at 1Hz

        // This sets the buzzer frequency
        if (ctrl_struct.alarming)
            ICR1 = 250;
        else 
            ICR1 = 150;
        OCR1A = OCR1B = ICR1/2;
        // ok alarm is ringing!
	if (effect_ctrl.beeping & 0x01) 
            { // first bit indicates pulsing alarm state
            effect_ctrl.beeping &= ~0x01;
            TCCR1B &= ~_BV(CS11); // turn buzzer off!
            PORTC &= ~_BV(LED1);
            } 
        else 
            {
            effect_ctrl.beeping |= 0x01;
            TCCR1B |= _BV(CS11); // turn buzzer on!
            PORTC |= _BV(LED1);
            }
        }
}
